/* link.c

   Copyright (C) 1997
	Fabrice Popineau  <Fabrice.Popineau@supelec.fr>

   This file is part of Web2c for Win32.

   Web2c For Win32 is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version. 
   
   Web2c For Win32 is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with Web2c For Win32; if not, write to the Free Software
   Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*
  the goal of this file is to mimic Unix's ln command. This is a quick
  hack, in that it will only allow to remap to predefined .exe files.
  */

#include <io.h>
#include <fcntl.h>
#include <stdio.h>
#include <process.h>
#include <ctype.h>
#include <windows.h>

#define IS_DIRSEP(c) ((c) == '/' || (c) == '\\')

/* To be determined by 'dumpbin /symbols lnexe.exe' 
   web2c/Makefile automagically determines this value by making two passes
   when compiling lnexe.exe
   */

#ifndef RDATA
#include "lnexe.h"
#endif

#define marker "foobarbaz"

#undef DEBUG

/* Initialized with marker string */
const char newargv0[MAX_PATH] = marker;

/*
  This programme should emulate Unix's ln for .exe files under Win32.

  Usage: lnexe <src> <dest>

  Method used: this .exe file is located, copied to the destination
  using the new name for the destination, while changing the newargv0
  constant to the src file name. 

  Caveat: to be portable, the src file name should not be absolute. 
  It should be either a pathname relative to the current executable file,
  either a simple name which may be looked for along the PATH, or only 
  in the same directory as the current executable file.

  Adopted strategy: relative pathes are ok, absolute one will give a 
  warning when doing the link.

  */

char cmdline[MAX_PATH];
char autoloc[MAX_PATH];
char *progname;

BOOL WINAPI catch_interrupt(DWORD arg)
{
  switch (arg) {
  case CTRL_C_EVENT:
  case CTRL_BREAK_EVENT:
    return TRUE;
  default:
    /* No need to set interrupt as we are exiting anyway */
    return FALSE;
  }
}

is_absolute(const char *path)
{
  return IS_DIRSEP(*path)
    || ( isalpha(*path) && *(path+1) == ':' && IS_DIRSEP(*path) );
}

is_relative(const char *path)
{
  return (*path == '.' && 
	  (IS_DIRSEP(*(path+1))
	   || (*(path+1) == '.' && IS_DIRSEP(*(path+2)))));
}

int strcasecmp(char *s1, char *s2)
{
  for (; *s1 && *s2 && toupper(*s1) == toupper(*s2); s1++, s2++);
  return (*s1 - *s2);
}

int link(char *src, char *dest, int force, int verbose)
{
  int fdest;

  if (verbose)
    fprintf(stdout, "Linking %s to %s.\n", src, dest);

  if (is_absolute(src))
    fprintf(stderr, "Warning: source file %s is absolute!\n", src);

  /* Looking for the source. */
  if (SearchPath(NULL, "lnexe", ".exe", MAX_PATH, autoloc,
		 NULL) == 0) {
    fprintf(stderr, "lnexe: can't find the program.\n");
    exit(1);
  }

#ifdef DEBUG
  fprintf(stderr, "Autoloc = %s\n", autoloc);
#endif

  if (CopyFile(autoloc, dest, !force) == 0) {
    fprintf(stderr, "Can't copy to %s (file may already exist).\n",
	    dest);
    exit(1);
  }
  /* Looking for the target. */
#if 0
  /* No need to look for src, which may be relative */
  if (SearchPath(NULL, src, ".exe", MAX_PATH, cmdline,
		 NULL) == 0) {
    fprintf(stderr, "%s: can't find the target %s.\n",
	    progname, src);
    exit(1);
  }
#endif

#ifdef DEBUG
  fprintf(stderr, "Cmdline = %s\n", cmdline);
#endif

  /* Now we need to put cmdline into the newargv0 variable in dest
     file.
     We could use the PE file format specfication to locate the .rdata
     section , but is it actually needed :-)
     In fact, we need to find the marker string "foobarbaz" and store
     the name here.
     */
  /* Open the file */
  if ((fdest = _open(dest, _O_RDWR | _O_BINARY)) == 0) {
    perror("Can't open target file");
    exit(1);
  }
  /* Look for the marker string */
  if (_lseek(fdest, RDATA, SEEK_SET) < 0) {
    perror("Can't seek to .rdata");
    exit(1);
  }

  if (_read(fdest, autoloc, MAX_PATH) < 0) {
    perror("Can't read into dest file");
    exit(1);
  }
  if (strcmp(autoloc, newargv0) != 0) {
    fprintf(stderr, "lnexe: marker string not found in %s.\n", dest);
    exit(1);
  }
    
  if (_lseek(fdest, RDATA, SEEK_SET) < 0) {
    perror("Can't seek (2) to .rdata");
    exit(1);
  }
    
  if (_write(fdest, src, strlen(src)+1) < 1) {
    perror("Can't write into dest file");
    exit(1);
  }

  /* Close the file */
  _close(fdest);

  return 0;
}

void usage()
{
      fprintf(stderr, "Usage: %s [-f] [-v] <src> <dest>\nwill link <src> to <dest>\n",
	      progname);
      fprintf(stderr, "\tOption -f will force the link, -v is verbose.\n");
      fprintf(stderr, "\tIf src is absolute, a warning is emitted because the hard link is not portable.\n");
      fprintf(stderr, "\tIf src is a relative path, it is relative to the dest path.\n");
      exit(1);
}

int
main (int	argc,
      char **	argv)

{
  int rc;
  size_t j = 0;
  int force = 0, verbose = 0;

  for (progname = argv[0] + strlen(argv[0]) -1;
       progname > argv[0] && *(progname-1) != '/' && *(progname-1) != '\\';
       progname--);
#ifdef DEBUG
  fprintf(stderr, "progname = %s\n", progname);
#endif

#ifdef DEBUG
  fprintf(stderr, "argc = %d, argv[0] = %s\n", argc, argv[0]);
#endif

  if (strcasecmp(progname, "lnexe.exe") == 0
      || strcasecmp(progname, "lnexe") == 0) {
    /* Doing the link */
    
    if (argc == 1)
      usage();

    /* First, parsing arguments */
    while (argc--, *(++argv)[0] == '-') {
#ifdef DEBUG
      fprintf(stderr, "arg: %s\n", argv[0]);
#endif
      switch (*(argv[0]+1)) {
      case 'f':
	force = 1;
	break;
      case 'v':
	verbose = 1;
	break;
      }
    }

    if (argc != 2)
      usage();
    link(argv[0], argv[1], force, verbose);
  }
  else {
    /* Callink the target program. Unfortunately, exec does not seem
       to work under NT as expected. */

    SetConsoleCtrlHandler(catch_interrupt, TRUE);

    if (is_relative(newargv0)) {
      /* We will catenate the directory where the link lies
	 to newargv0 */
      char *fp;
      if (SearchPath(NULL, progname, ".exe", MAX_PATH, autoloc,
		     &fp) == 0) {
	fprintf(stderr, "%s: can't autolocate!\n", progname);
	exit(1);
      }
      strcpy(fp, newargv0);
#ifdef DEBUG
      fprintf(stderr, "Relative catenated = %s\n", autoloc);
#endif
    }
    else
      strcpy(autoloc, newargv0);

#ifdef DEBUG
    fprintf(stderr, "Launched %s\n", autoloc);
#endif

    rc = _spawnvp (_P_WAIT, autoloc, argv);
    
    if (rc < 0) {
      perror (cmdline);
      return (1);
    }
    
    return (rc);
  }
  return 0;
}

/* link.c ends here */
