#
# Build tool for Web2C & supp
# Author: F. Popineau
# Date  : 04/11/98
# Time-stamp: <02/02/28 21:28:43 popineau>
#
# Usage: 
#        build <option>
#              [--static]         builds/cleans up statically
#              [--debug]          builds/cleans up in debug mode
#              [--depend]         updates dependencies
#              [--clean]          cleans up 
#              [--distclean]      removes everything that can be rebuild (needs to configure again)
#              [--dir <dir>]      only in <dir> (relative to top !)
#              [--install]        installs what has been built (incompaible with clean up targets)
#              [--from-scratch]   full build, removes everything from the destination tree.
#              [--log <file>]     logs everything onto <file>
#
#        Hope these options are clear enough. This is a wrapper to nmake calls.
#

require "newgetopt.pl";
use lib "./perl";
use file_utils qw(canon_dir cleandir make_link newpath 
		  normalize substitute_var_val dirname
		  rec_rmdir sync_dir walk_dir start_redirection
		  stop_redirection sync_dir);
use tpm qw(Tpm::Tpm2Zip Tpm::Pkg2Zip);
use Cwd;
use Time::gmtime;

# list if functions to be called before nmake is run
@prepare_func_list = ( );

$fptexversion = "0.5-beta3";
$Gm = gmtime();
$Date = sprintf("%02d-%02d-%02d", $Gm->year() + 1900, $Gm->mon(), $Gm->wday());
# root directory
$topdir = dirname(cwd());
$win32_dir = "$topdir/win32";
$win32_mak_dir = "$win32_dir/make";
$pkgbindir = "$win32_dir/factory/pkg/pkg-bin";
$zipsrc_name = "d:/InetPub/ftp/fptex/sources/fptex-$fptexversion-$Date.zip";
$masterdir="d:/source/TeXLive/Master";
$destdir="d:/Local/TeXLive";
$logfile = "";

@pkg_standalone = ( "pdftex", "omega", "dvipdfm", "windvi", "gsftopk", "tex4htk" #, "ltx2rtf"
		  );
@src_standalone = (
		   "texk/kpathsea", "texk/contrib", "texk/web2c",
		   "texk/dvipdfm", "texk/windvi",
		   "texk/odvipsk", "texk/owindvi",
		   "texk/gsftopk", "texk/tex4htk"
#		   "texk/ltx2rtf"
		  );

# list of pathes where to look for a tetex-texmf.tar.gz file
@tetex_texmf_path = ( "$topdir\\..\\..\\Mirror\\TeX\\teTeX-beta",
		      "$topdir\\..\\..\\Mirror\\TeX",
		      "$topdir\\..\\Mirror\\TeX\\teTeX-beta",
		      "$topdir\\..\\Mirror\\TeX",
		      "$topdir\\..\\..\\orig\\teTeX-beta", 
		      "$topdir\\..\\..\\orig", 
		      "$topdir\\..\\orig\\teTeX-beta", 
		      "$topdir\\..\\orig",
		      "$topdir\\..\\teTeX-beta", 
		      "$topdir\\.."
		    );

# list of directories in $(prefix) that will be created by nmake
@fptex_dirs = ( "bin", "include", "lib", "setupw32" );

@zipsrc_exclude_dirs = (
			"/(dynamic|dynamic-debug|static|static-debug|Release|Debug)\$",
			"/win32/(diff|setupw32|factory|ttfinst|worked-on)\$",
			"/win32/(deutsch|english|fptex.old|french|labels|slovak|tds)\$",
			"/utils/rxp-1.2\$",
			"/texk/(ltx2rtf-old|windvi-0.68|texdoctk-0.3.0|texdoctk-0.4.0)\$"
		       );

@zipsrc_exclude_patterns = (
 "\.(obj|dll|exe|opt)\$",
 "~\$"
);


&main;

1;

sub help
{
    print "Usage: build [options]\n";
    print "\n";
    print "    --dir <path>     Apply to the subdir <path> (relative);\n";
    print "    --log <file>     Redirect stdin and stderr in <file>\n";
    print "    --static         Build statically linked versions;\n";
    print "    --debug          Build debug versions;\n";
    print "    --sync           Sync up with master;\n";
    print "    --standalone     Build standalone packages (ie: static, nodebug);\n";
    print "\n";		       
    print "    Targets :\n";   
    print "      <none>         Compiles;\n";
    print "    --clean          Clean up last build (only compilation);\n";
    print "    --distclean      Clean up last build (everything, must configure again);\n";
    print "    --build          Build everything;\n";
    print "    --install        Install everything build;\n";
    print "    These targets are exclusive !\n";
    print "\n";
    print "    --from_scratch   Use with --install and --build only,";
    print "                     will remove everything from <prefix> directory,\n";
    print "    --new_texmf      Use with --from_scratch, will install\n";
    print "                     a fresh copy of teTeX texmf tree;";
    print "    --prefix         Used with --install only,";
    print "                     will change temporarily the <prefix> directory,\n"; 
    print "\n";
    print "    Support commands:\n";
    print "    --help           displays this help text and exits\n";
    print "    --verbose        some more messages\n";
}

sub read_config_var
{
    local($file, $varname) = @_;
    local($val);
    open VER, "<$file";
    while (<VER>) {
      if ($_ =~ m/^$varname\s*=\s*(.*)$/) {
	$val = $1;
      }
    }
    close VER;
    return $val;
}

sub make_absolute
{
  local ($rel) = @_;
  local ($current, $abs);
  $current = cwd();
  chdir $rel;
  $abs = cwd();
  print "current = $current, rel = $rel, abs = $abs\n";
  chdir $current;
  return $abs;
}

sub from_scratch
{
  local ($d, $prefix, $l, $highest, $srcdir, $dstdir);

  $prefix = &read_config_var("$win32_mak_dir/paths.mak", "prefix");
  $prefix = &canon_dir($prefix, '\\');
  print "Prefix directory is $prefix .\n" if $opt_verbose;

  # Removing every fpTeX thing from prefix directory
  foreach $d ( @fptex_dirs ) {
    if (-d "$prefix\\$d") {
      &cleandir("$prefix\\$d");
      rmdir "$prefix\\$d";
    }
  }

  return if (! $opt_new_texmf);

  # Look around for teTeX-texmf trees 
  for $d ( @tetex_texmf_path ) {
    print "directory $d\n";
    print "files " . <$d\\teTeX-texmf-*.tar.gz> . "\n";
    push @l, <${d}\\teTeX-texmf-*.tar.gz>;
  }

  if ($#l == -1) {
      print "No teTeX-texmf-*.tar.gz file found.\nCheck build.pl for search path.\n";
      return;
  }

  @l = sort @l;
  $highest = $l[$#l];

  for $d ( @tetex_texmf_path ) {
    if (-f "$d\\$highest") {
      $highest = &make_absolute($d) . "\\$highest" ;
      break;
    }
  }
  print "using $highest as texmf tree.\n" if $opt_verbose;

  open TEXMF, "<$prefix\\texmf\\stamp-texmf";
  while (<TEXMF>) {
    $current = $_;
  }
  close TEXMF;

  if ($current ge $highest) {
    print "Current texmf tree is $current and no other one is more recent ($highest)\n";
    return;
  }

  print "Current texmf tree is $current and $highest is more recent.\n";
  print "Current tree will be removed and replaced.\n";

  # removing current texmf tree
  &rec_rmdir("$prefix\\texmf");

  # Installing a fresh copy of texmf
  $current = cwd();
  mkdir "$prefix\\texmf", "666";
  chdir "$prefix\\texmf";
  `gzip -dc $highest | tar xvf -`;
  @l = `gzip -dc $highest | tar tvf -`;
  for $_ (@l) {
    if ($_ =~ m/([^\s]+)\s*->\s*([^\s]+)/) {
      print "Linking $1 -> $2\n" if $opt_verbose;
      make_link($1, $2);
    }
  }

  # and stamping it
  open TEXMF,">stamp-texmf";
  print TEXMF "$highest";
  close TEXMF;

  chdir $current;
}

sub mask_env {
  $ENV{"SystemDrive"} = "foo";
  $ENV{"TEXMFCNF"} = "foo";
  $ENV{"TEXMFMAIN"} = "foo";
  $ENV{"TEXMF"} = "foo";
  $ENV{"LATEX2HTML"} = "foo";
  $ENV{"GROFF_TMAC_PATH"}="foo";
  $ENV{"GROFF_FONT_PATH"}="foo";
}

sub ziplist {
  my ($dir, @l) = @_;
  $dir =~ s@\\@/@g;
  map {
    if (! -d "$dir/$_") {
      for $pattern (@zipsrc_exclude_patterns) {
	if ("$dir/$_" =~ m@$pattern@i) {
	  goto suivant;
	}
      }
      print STDERR "$dir/$_\n" if ($opt_verbose);
      print ZIP "$dir/$_\n";
    }
  suivant:
  } @l;
}

sub zipprune {
  my ($dir) = @_;
  $dir =~ s@\\@/@g;
  map {
    if ($dir =~ m@$_@) {
      print STDERR "Pruning $dir...\n" if ($opt_verbose);
      return 1;
    }
  } @zipsrc_exclude_dirs;
  return 0;
}

sub do_zipsrc {
  unlink $zipsrc_name if (-f $zipsrc_name);
  open ZIP, "| zip -9\@or " . $zipsrc_name . " > nul";
  &walk_dir($topdir, \&ziplist, \&zipprune);
  close ZIP;
}

sub do_standalone {
  local($dyn_dir, $static_dir);
  local($cmd, $cwd);

  my $dyn_dir = "$prefix/bin/win32-orig";
  my $static_dir = "$prefix/bin/win32";
  my $tmp_dir = "$prefix/bin/win32-tmp";

  # move the bin directory to some safe location
  if (! -d $static_dir) {
    if (-d $dyn_dir) {
      # Assume nothing to do
    }
    else {
      print STDERR "!!! No $static_dir nor $dyn_dir";
    }
  }
  else {
    rename ($dyn_dir, $tmp_dir);
    rename ($static_dir, $dyn_dir);
    &rec_rmdir($tmp_dir);
  }

  # make the static binaries needed
  print STDERR "Installing in $prefix\n";

  &mask_env;

  &start_redirection($logfile);

  for $d (@src_standalone) {
    my $cwd = &getcwd;
    print "Changing directory to $topdir/$d\n" if $opt_verbose;
    chdir("$topdir/$d");
    $cmd = "nmake -f win32.mak NODEBUG=1 STATIC=1 install";
    print "Running *twice* $cmd\n" if $opt_verbose;
    system($cmd);
    system($cmd);
    chdir($cwd)
  }

  # build the zip files
  for $d (@pkg_standalone) {
    my $pkg_name = "$pkgbindir/$d-static.pkg";
    if (! -f "$pkg_name") {
      $pkg_name = "$pkgbindir/$d.pkg";
      if (! -f $pkg_name) {
	printf STDERR "Non existent package description for $d ... skipping!\n";
	next;
      }
    }
    &Tpm::Pkg2Zip($pkg_name, $prefix, "d:/InetPub/ftp/fptex/standalone");
  }

  &stop_redirection($logfile);

  # clean up
  &rec_rmdir($static_dir);
  # put binaries back in place
  rename($dyn_dir, $static_dir);
}

sub do_sync {

  $cmd = "p4 sync";
  system($cmd);

  &sync_dir("$masterdir/texmf", "$destdir/texmf");
  &sync_dir("$masterdir/texmf/web2c", "$destdir/texmf-var/web2c");

}

sub main {

  local ($changed_prefix);

  unless (&NGetOpt ("static", "debug", 
		    "clean", "distclean" ,
		    "depend","from_scratch", "new_texmf",
		    "build", "install",
		    "zipsrc:s",
		    "standalone", "sync",
		    "dir=s", "log=s", "prefix=s",
		    "verbose", "help"))
    {
	print STDERR "Try `configure --help'";
	exit 1;
    }

  if ($opt_help) {
    &help;
    exit 0;
  }

  $prefix = &read_config_var("$win32_mak_dir/paths.mak", "prefix");

  if ($opt_log) {
    $logfile = "$topdir\\win32\\$opt_log";
    if (! $opt_log =~ /\.log$/) {
      $opt_log .= ".log";
    }
  }

  if ($opt_standalone) {
    &do_standalone;
    exit 0;
  }

  if (defined($opt_zipsrc) && $opt_zipsrc ne "clean") {
    &do_zipsrc;
    exit 0;
  }
  if (defined($opt_zipsrc) && $opt_zipsrc eq "clean") {
    $opt_distclean = 1;
  }

  # build the command line
  $prog = "nmake -f win32.mak ";
  $prog .= "verbose=on " if ($opt_verbose);
  if ($opt_debug eq '') {
    $prog .= "NODEBUG=1 ";
  }
  if ($opt_static) {
    $prog .= "STATIC=1 ";
  }
  if ($opt_depend) {
    $prog .= "MAINT=1 ";
    $target = "depend ";
  }
  if ($opt_build) {
    if ($target ne '') {
      print "Conflicting targets : $target and build\n";
      return 1;
    }
    $target = "all ";
    if ($opt_from_scratch) {
	push @prepare_func_list , ( \&from_scratch );
    }
  }
  if ($opt_install) {
    if ($target ne '') {
      print "Conflicting targets : $target and install\n";
      return 1;
    }
    $target = "install ";
    if ($opt_from_scratch) {
	push @prepare_func_list , ( \&from_scratch );
    }
  }
  if ($opt_clean) {
    if ($target ne '') {
      print "Conflicting targets : $target and clean\n";
      return 1;
    }
    $target = "clean ";
  }
  if ($opt_distclean) {
    if ($target ne '') {
      print "Conflicting targets : $target and distclean\n";
      return 1;
    }
    $prog .= "MAINT=1 ";
    $target = "distclean ";
  }

  if ($target eq '') {
      print "No target specified.\nEnd.\n";
      exit 0;
  }

  if ($opt_prefix) {
    if ($target ne "install ") {
      print STDERR "$0: --prefix option ignored, wrong target.\n";
    }
    elsif (! -d $opt_prefix) {
      print STDERR "$0: $opt_prefix is not a directory, aborting.\n";
      exit 1;
    }
    else {
      $opt_prefix = &normalize($opt_prefix);
      if ($prefix ne $opt_prefix) {
	print STDERR "Replacing prefix $prefix with $opt_prefix.\n";
	@old_prefix = &substitute_var_val("$win32_mak_dir/paths.mak", 
					  "prefix", "$opt_prefix");
	$changed_prefix = 1;
	$old_prefix = $prefix;
	$prefix = $opt_prefix;
      }
    }
  }

  print STDERR "Installing in $prefix\n";

  &mask_env;

  $prog .= $target;

  $buildir = $topdir;

  if ($opt_dir) {
    $buildir .= "\\$opt_dir";
  }
  if (-d $buildir) {
	chdir $buildir;
  }
  else {
	print "Non-existent dir: $opt_dir\n";
	return 1;
  }

  # start redirection if asked
  &start_redirection($logfile);

  if ($opt_sync) {
    &do_sync;
  }

  # Prepare functions
  for $f (@prepare_func_list) {
    if ($f) {
      &{$f};
    }
  }

  # Call the program
  #`$prog`;
  print STDOUT "Executing $prog in " . &getcwd . "\n";
  system($prog);

  # stop redirection
  stop_redirection($logfile);

  chdir $topdir;

  if ($#old_prefix > -1) {
    &substitute_var_val("$win32_mak_dir/paths.mk", 
			"prefix", shift @old_prefix);
  }

  if ($opt_distclean && $opt_zipsrc) {
    &do_zipsrc;
  }

  print "This is the end!\n";
}
