package file_utils;

BEGIN {
  use Exporter ();
  use Cwd;
#  use File::Copy qw(copy);
  use vars qw( @ISA @EXPORT_OK);
  @ISA = qw(Exporter);

  @EXPORT_OK = qw(
		  &basename
		  &build_path
		  &build_tree
		  &calc_file_size
		  &canon_dir
		  &check_path
		  &cleandir
		  &set_file_time
		  &copy
		  &diff_list
		  &dirname
		  &globexpand
		  &is_absolute
		  &is_dirsep
		  &look_for
		  &make_link
		  &member
		  &min max
		  &move
		  &newer
		  &newpath
		  &normalize
		  &print_tree
		  &push_uniq
		  &rec_copy
		  &rec_mkdir
		  &rec_rmdir
		  &remove_list
		  &sort_uniq
		  &start_redirection
		  &stop_redirection
		  &substitute_var_val
		  &sync_dir
		  &walk_dir
		  &walk_tree
		 );
}

#
# For those functions that use recursion
#

$_recurse = 1;

# Is the character a directory separator ?
sub is_dirsep
{
    my ($c) = @_;
    if ($c =~ /[\/\\]/) {
	return 1;
    }
    else {
	return 0;
    }
}

# Is the path absolute ?
sub is_absolute
{
    my ($d) = @_;
    if ($d =~ m@^([A-Za-z]:)?[/\\]@) {
      return 1;
    }
    else {
      return 0;
    }
}

# Rewrite '\\' into '/' and deletes multiple ones/
sub canon_dir
{
    my ($p, $rep) = @_;
    if ($p =~ m@^(.*)[/\\]$@) {
      $p = $1;
    }
    if (($rep eq '') || ($rep eq '\\')) {
	$p =~ s@/@\\@g;
	$p =~ s@\\[\\]+@\\@g;
	$p =~ s@\\\.\\@\\@g;
	$p =~ s@\\([^\\]+)\\..\\@\\@g;
    }
    elsif ($rep eq '/') {
	$p =~ s@\\@/@g;
	$p =~ s@/[/]+@/@g;
	$p =~ s@/\./@/@g;
	$p =~ s@/([^/]+)/../@/@g;
    }
    else {
	die ("canon_dir($p) : invalid separator $rep.\n");
    }
    return $p;
}

# Merges all elements in the list into a single path, adding
# directory separators as needed.
sub build_path
{
    my($p, $s);
    # Concatenates the arguments, adding path separators as needed
    $p = $_[0];
    for ($i = 1; $i <= $#_; $i++) {
	$p = $p . "\\" . $_[$i];
    }
    return &canon_dir($p);
}

sub sort_uniq {
  my (@l) = @_;
  my ($e, $f, @r);
  @l = sort(@l);
  foreach $e (@l) {
    if ($e ne $f) {
      $f = $e;
      push @r, $e;
    }
  }
  return @r;
}

sub remove_list {
  local (*l, $e) = @_;
  my (@r, $f);
  foreach $f (@l) {
    if ($e ne $f) { push @r, $f; }
  }
  @l = @r;
}

sub member {
    my ($e, @l) = @_;
    my ($f);
    foreach $f (@l) {
	if ($e eq $f) {
	    return 1;
	}
    }
    return 0;
}

sub push_uniq {
  local (*l, @le) = @_;
  my ($e);
  foreach $e (@le) {
    if (! &member($e, @l)) {
      push @l, $e;
    }
  }
}

sub dirname
{
    my ($f) = @_;
    $f =~ m@(^.*)[\\/][^\\/]*$@;
    return $1;
}

sub basename
{
    my ($f) = @_;
    $f =~ m@([^\\/]*)$@;
    return $1;
}

sub normalize
{
    my ($p, $sep) = @_;
    if ($sep eq '/') {
      $p =~ s@\\@/@g;
      $p =~ s@/(/|\./)*@/@g;
      return $p;
    }
    elsif ((! $sep) || ($sep eq '\\')) {
      $p =~ s@/@\\@g;
      $p =~ s@\\(\\|\.\\)*@\\@g;
      return $p;
    }
    else {
      print STDERR "normalize : invalid separator, $sep\n";
      return $p;
    }
}

sub walk_dir
{
    # Walks the directory, executing $proc for each file,
    # until done is returned.
    my ($dir, $proc, $prune) = @_;
    my (@l, $f, $done, $src, $DIR);

    print "Walking $dir\n" if $opt_verbose;

    if ((! $prune) || ($prune && ! &{$prune}($dir))) {
	$done = 0;
	# Walk the directory tree
	opendir DIR, "$dir";
	# do not forget to remove "." and ".."
	@l = readdir (DIR); shift @l; shift @l;
	closedir DIR;
	# top-down
	&{$proc}($dir, @l);
	foreach $f (@l) {
	    if (-d $dir . "\\" . $f) {
		&walk_dir($dir . "\\" . $f, $proc, $prune);
	    }
	}
    }
}

# Builds up a tree from pathes
# $node is a reference to a hash
sub build_tree {
  my ($node, @elts) = @_;
  my ($current);

  while (@elts) {
    $current = shift @elts;
    if ($$node{$current}) {
      $node = $$node{$current};
    }
    else {
      $$node{$current} = { };
      $node = $$node{$current};
    }
  }
}

# walks the tree, calling the function at each node
sub walk_tree {
  my(@stack_dir);
  walk_tree1(@_);
}

sub walk_tree1 {
  my ($node, $proc) = @_;
  &{$proc}($node, @stack_dir);

  for $k (keys(%{$node})) {
    push @stack_dir, $k;
    $v = $node->{$k};
    walk_tree1 (\%{$v}, $proc);
    pop @stack_dir;
  }
}

# Check that a path exists in a tree
# exactly or as a subpath
sub check_path
{
  my ($node, @path, $exact) = @_;
  my ($current);
#  print "Checking for " . join('/', @path) . " exact = $exact\n";
  while (@path) {
    $current = shift @path;
    if ($$node{$current} == undef) {
      return 0;
    }
    else {
      $node = $$node{$current};
    }
  }
#  print "left " .  join('/', @path) . " leaf " .  $#{keys %{$node}} . "\n";
  if ($exact) {
    # We are at a leaf !
    return ($#{keys %{$node}} == -1);
  }
  else {
    return 1;
  }
}



# Removes everything in the directory
# Can't use walk_dir because it could not remove directories.
sub cleandir {
  my ($dir)= @_;
  my ($DIR, $f, @l, $name);

  if (-d $dir) {
    opendir DIR, "$dir";
    # do not forget to remove "." and ".."
    @l = readdir (DIR); shift @l; shift @l;
    closedir DIR;
    foreach $f (@l) {
      $name = $dir . "\\" . $f;
      if (-d $name) {
	&rec_rmdir($name);
	print "rmdir $name\n" if $opt_verbose;
	rmdir "$name";
      }
      elsif (-f "$name") {
	print "Removing $name\n" if $opt_verbose;
	unlink "$name";
      }
      else {
	print "Can't remove $name!\n";
      }
    }
  }
}

sub rec_rmdir {
  my (@files) = @_;
  map { &cleandir($_); rmdir($_) } @files;
}

# Builds up a new directory together with any of its parents
sub rec_mkdir {
  my ($path) = @_;
  my ($l, $tmp, $tmppath, $dir);
  $dir = &getcwd;
  print "mkdir $path\n";
  @l = split ("[/\\\\]", $path);
  if ($l[0] =~ m/[A-Za-z]:/) {
    chdir "$l[0]/";
    shift @l;
  }
  while (@l) {
    $tmp = shift(@l);
    mkdir($tmp, 777 & !umask) if (! -d $tmp);
    chdir $tmp;
  }
  chdir $dir;
}

sub copy {
 my (@src, $dest, $l, $t);
 @l = @_;
 $dest = pop @l;

 @src = ();
 map { push @src, <${_}>; } @l;

 map {
   if (! -d $_) {
     open IN, "<$_";
     if (-d $dest) {
	open OUT, ">" . &newpath($dest, &basename($_));
     }
     else {
	open OUT, ">$dest";
     }
     binmode(IN);
     binmode(OUT);
     print OUT <IN>;
     close(OUT);
     close(IN);
     @t = stat($_);
     utime($t[8], $t[9], &newpath($dest, &basename($_)));
   }
 } @src;
}

sub set_file_time {
  my ($from, $to) = @_;
  @st = stat($from);
  utime($st[8], $st[9], $to);
}

sub rec_copy {
  my (@src, $dest, $l, $dir);
  @l = @_;
  $dest = pop @l;

  @src = ();
  map { push @src, <${_}>; } @l;

  if (! -d $dest) {
    mkdir ($dest, 777);
  }

  $dir = &getcwd;
  if (! is_absolute($dest) ) {
    $dest = &canon_dir("$dir/$dest");
    print "new dest = $dest\n";
  }

  map {
    $l = $_;
    if (-d $l) {
      chdir($l);
      &rec_copy("*", "$dest/" . &basename($l));
      chdir($dir);
    }
    else {
      &copy($l, "$dest/" . &basename($l));
      &set_file_time($l, "$dest/" . &basename($l));
    }
  }  @src;
}

sub move {
  my ($dest) = pop @_;
  my ($src, $f, $l);

  @l = @_;
  foreach $f (@l) {
    # Handle globbing
    if ($f =~ m/[*?]/) {
      push @src, <${f}>;
    }
    else {
      push @src, $f;
    }
  }
  if (($#src > 2) && (! -d $dest)) {
    print STDERR "*** Move : can't move to $dest, not a directory.\n";
    return;
  }

  foreach $f (@src) {
    if (-d $dest) {
      rename($f, &newpath($dest, &basename($f)));
    }
    else {
      &copy($f, $dest);
      &set_file_time($f, $dest);
      unlink($f);
    }
  }
}

# Simulate links by copying
sub make_link {
    my ($to, $from) = @_;
    $to = canon_dir($to);
    $from = &newpath(dirname($to), $from);
    print "linking $from -> $to ...";
    if (-e $to) {
	unlink($to);
    }
    if (-d $from) {
      system("xcopy $from $to /f/r/i/e/d/k");
    }
    else {
	&copy($from, $to);
	&set_file_time($from, $to);
    }
    print " done\n";
}

# Merges all elements in the list into a single path, adding
# directory separators as needed.
sub newpath
{
  return &canon_dir(join ("\\", @_));
}

#
# Search for $key = $val in $file
#
sub look_for
{
    my($key, $file) = @_;
    my($ret);
    open FIN, "<$file";
    while (<FIN>) {
	if ($_ =~ m/^$key\s*=\s*(\S*)/) {
	    $ret = $1;
	    last;
	}
	elsif (/^\#define\s+$key\s+(\S*)/) {
	  $ret = $1;
	  last;
	}
    }
    close FIN;
    return $ret;
}


sub max {
  $m = shift;
  while ($_ = shift) {
    $m = $_ if ($_ > $m);
  }
  return $m;
}

sub min {
  $m = shift;
  while ($_ = shift) {
    $m = $_ if ($_ < $m);
  }
  return $m;
}

# Changes lines of the form:
# $var=... to
# $var='$val' in $file
#
sub substitute_var_val {
    my($file, $var, $val) = @_;
    my($success);

    @success = ( );

    if (! -f $file) {
      print STDERR "$0: $file is not a file\n";
      return $success;
    }
    open IN, "<$file";
    open OUT, ">$file.bak";

    while (<IN>) {
      s/^$var\s*=\s*(.*)$/do {
	push @success, $1;
	#		   "\$" . $var . "=" . eval('$val') . ";" .
	"$var = $val"
      }/e;
      print OUT;
    }
    close IN;
    close OUT;

    &copy("$file.bak", "$file");
    unlink("$file.bak");

    return @success;
}

#
# Used by globexpand($recurse, $dir)
#
sub globexpand_push {
  my ($dir, @l) = @_;
  my ($file);
  $dir =~ s@\\@/@g;
  foreach $file (@l) {
    if (-f "$dir/$file") {
      push @listglob, "$dir/$file";
    }
  }
}

#
# Returns the list of files that match $pattern
# Recursively walking directories if $recurse
#
sub globexpand {
  my ($recurse, $pattern) = @_;
  local @listglob = ( );
  while ( <${pattern}> ) {
    print "elt $_\n" if $opt_verbose;
    if (-f $_) {
      push @listglob, $_;
    }
    elsif (($_ ne "") && (-d $_) && ($recurse)) {
      &walk_dir($_, \&globexpand_push);
    }
  }
  print "list = $pattern => " . @listglob . "\n" if $opt_verbose;
  return @listglob;
}

sub diff_list {
  local ($l1, $l2, *l1_l2, *l2_l1) = @_;

#   print "Before sorting:\n";
#   map { print "$_\n"; } @$l1;
#   print "\n";
#   map { print "$_\n"; } @$l2;
#  $opt_verbose = 1;
#  $opt_debug = 1;

  @l1 = sort(@$l1);
  @l2 = sort(@$l2);

#   print "After sorting:\n";
#   map { print "$_\n"; } @l1;
#   print "\n";
#   map { print "$_\n"; } @l2;

  while ($#l1 >= 0 || $#l2 >= 0) {
    my $comp = $l1[0] cmp $l2[0];
    if ($#l1 == -1) {
      print "No more elements in l1, over.\n" if $opt_debug;
      push (@l2_l1, @l2);
      @l2 = ();
    }
    elsif ($#l2 == -1) {
      print "No more elements in l2, over.\n" if $opt_debug;
      push (@l1_l2, @l1);
      @l1 = ();
    }
    elsif ($comp == 0) {
      print "Same element $l1[0], shifting both.\n" if $opt_debug;
      shift @l1;
      shift @l2;
    }
    elsif ($comp > 0) {
      print "Greater element $l1[0] than $l2[0], shifting l2.\n" if $opt_debug;
      push (@l2_l1, shift @l2);
    }
    else {
      print "Smaller element $l1[0] than $l2[0], shifting l1.\n" if $opt_debug;
      push (@l1_l2, shift @l1);
    }
  }

  if ($opt_verbose) {
    print "$#l1_l2, $#l2_l1\n";

    print "Elts in l1 and not in l2 : \n";
    map { print "$_\n"; } @l1_l2; print "\n";

    print "Elts in l2 and not in l1 : \n";
    map { print "$_\n"; } @l2_l1; print "\n";
  }
}

sub sync_dir {
  my ($src, $dst, $opt_proc, $opt_prune, $opt_dry, $opt_mirror) = @_;
  local ($cwd, $dry, $mirror, $proc, $prune);

  $cwd = &getcwd;
  $dry = $opt_dry;
  $mirror = $opt_mirror;
  $proc = $opt_proc;
  $prune = $opt_prune;
  chdir($src);
  sync_dir_1(".", $dst);
  chdir($cwd);
}

sub newer {
  my ($f1, $f2) = @_;
  my ($t1, $t2, $t11, $t12);
  @t1 = stat($f1);
  @t2 = stat($f2);
  $t11 = $t1[9];
  $t12 = $t2[9];
  return -1 if ($t11 < $t12);
  return 1 if ($t11 > $t12);
  return 0;
}

sub sync_dir_1
{
  # Walks the directory, executing $proc for each file,
  # until done is returned.
  my ($dir, $dst) = @_;
  my (@l, $f, $done, $src, $DIR);

  print "Walking $dir\n" if $opt_verbose;

  if (! -d $dst) {
    if (-f $dst) {
      # Clash
      print "!!!Clash: $dir is a directory and $dst is a file.\n";
      return;
    }
    else {
      print "Creating missing directory $dst\n";
      mkdir $dst if (! $dry);
    }
  }

  if ((! $prune) || ($prune && ! &{$prune}($dir))) {
    $done = 0;
    # Walk the directory tree
    opendir DIR, "$dir";
    # do not forget to remove "." and ".."
    @l = readdir (DIR); shift @l; shift @l;
    closedir DIR;
    # top-down
    &{$proc}($dir, $dst, @l) if ($proc);
    foreach $f (@l) {
      if (-d $dir . "\\" . $f) {
	# source is directory
	&sync_dir_1($dir . "\\" . $f, $dst . "\\" . $f);
      }
      elsif (-d $dst . "\\" . $f) {
	# source is file and destination is directory
	print "!!!Clash: $dir is a file and $dst is a directory.\n";
	next;
      }
      elsif (! -f $dst . "\\" . $f) {
	# source is file and destination is missing
	print "Copying missing file " . $dst . "\\" . $f . "\n";
	if (! $dry) {
	  &copy ($dir . "\\" . $f, $dst . "\\" . $f);
	  &set_file_time($dir . "\\" . $f, $dst . "\\" . $f);
	}
      }
      else {
	my $compare = &newer($dir . "\\" . $f, $dst . "\\" . $f);
	if ($compare > 0 || ($mirror && $compare < 0)) {
	  # source is file and destination is older than source
	  print "Copying newer file " . $dir . "\\" . $f . " than " . $dst . "\\" . $f . "\n";
	  if (! $dry) {
	    &copy ($dir . "\\" . $f, $dst . "\\" . $f);
	    &set_file_time($dir . "\\" . $f, $dst . "\\" . $f);
	  }
	}
      }
    }
    # Look at the other side
    opendir DIR, "$dst";
    # do not forget to remove "." and ".."
    @l = readdir (DIR); shift @l; shift @l;
    closedir DIR;
    foreach $f (@l) {
      # We should look only for things to remove on the destination side
      next if (-e "$dir\\$f");
      # If it does not exist on the source side, then remove it.
      if (-d "$dst\\$f") {
	print "Removing directory $dst\\$f\n";
	&rec_rmdir("$dst\\$f") if (! $dry);
      }
      elsif (-f "$dst\\$f") {
	print "Removing file $dst\\$f\n";
	unlink("$dst\\$f") if (! $dry);
      }
      else {
	print "Unknown file type $f\n";
      }
    }

  }
}

sub start_redirection {
  local ($log) = @_;

  # start redirection if asked
  if ($log) {
    print "Logging onto $log\n";
    open(SO, ">&STDOUT");
    open(SE, ">&STDERR");

    close(STDOUT);
    close(STDERR);

    open(STDOUT, ">$log");
    open(STDERR,">&STDOUT");

    select(STDERR); $| = 1;
    select(STDOUT); $| = 1;
  }
}

sub stop_redirection {

  local($log) = @_;

  if ($log) {
      close(STDOUT);
      close(STDERR);
      open(STDOUT, ">&SO");
      open(STDERR, ">&SE");
  }
}

sub calc_file_size_proc {
  my ($d, @l) = @_;
  my @st = stat("$d/$_");
  map {
    $size += $st[7];
  } @l;
  return $size;
}

sub calc_file_size {
  my ($dir, @files) = @_;
  my ($size, @st);

  # print "calc_file_size : $dir, files = $#files\n"; # if $opt_debug;
  if (! -d $dir) {
    print STDERR "$0: $dir is not a directory!\n";
    return 0;
  }
  $size = 0;
  # no need to walk here, done by globexpand
  if (is_absolute($_)) {
    @files = map { &globexpand(1, "$_"); } @files;
  }
  else {
    @files = map { &globexpand(1, "$dir/$_"); } @files;
  }
  map {
    @st = stat($_);
    $size += $st[7];
  } @files;
  # print "size = $size\n" if $opt_debug;
  return $size;
}

END { }

1;
