#
# TPM Build tool for TeXLive
# Author: F. Popineau
# Date  : 27/03/2001
# Time-stamp: <02/05/31 10:56:06 popineau>
#

require "newgetopt.pl";
use lib ("./perl", "../perl");
use file_utils qw(canon_dir cleandir make_link newpath member
		  normalize substitute_var_val dirname diff_list
		  rec_rmdir sync_dir walk_dir start_redirection stop_redirection);
use tpm qw(Tpm::Pkg2Tpm Tpm::Pkg2Zip Tpm::Tpm2Zip 
	   Tpm::OldTpm2Pkg Tpm::extract_remote);
use Cwd;

use Strict;

$direction = 'pkg2tpm';
$opt_check = 0;
$opt_pkg_dir = "./factory/pkg";
$opt_tpm_dir = "./factory/texmf/tpm";
$opt_tex_dir = "${Tpm::masterdir}";
$opt_dest_dir = "${Tpm::ftpdir}/${Tpm::fptexversion}";
$opt_import_tpm_dir = "$opt_tex_dir/texmf/tpm";
$opt_arch = "win32";
$opt_log = 0;
$opt_debug = 0;
$opt_verbose = 0;

%srcdst = (
	   "pkg-scm" => "schemes",
	   "pkg-col" => "collections",
	   "pkg-dat" => "packages",
	   "pkg-bin" => "binaries",
	   "pkg-sup" => "support"
	  );

&main;

1;

sub help {
    print "Usage: ${ARGV[0]} [options]\n";
    print "\n";
    print "    --tpm2zip               build zip files out of a set of tpm files;\n";
    print "    --pkg2tpm               converts the factory pkg files to tpm files;\n";
    print "    --tpm_import            imports old tpm files to factory pkg files;\n";
    print "    --check=<(dep|cov)>     check dependencies or coverage for the tpm files.\n";
    print "    --tpm_index             rebuild the tpm.zip index file.\n";
    print "    These targets are exclusive !\n";
    print "\n";		       
    print "    Directories :\n";
    print "    --tex_dir=<path>        root directory of the TeX distribution.\n";
    print "    --tpm_dir=<path>        where to get the tpm files (\$opt_tex_dir/texmf/tpm by default).\n";
    print "    --dest_dir=<path>       destination directory for the zip files.\n";
#    print "    --subdir=<path>         Process the tpm files in subdirectory <path>.\n";
    print "    --pkg_dir=<path>        directory for pkg files;\n";
    print "    --tpm_dir=<path>        directory for tpm files;\n";
    print "    --import_tpm_dir=<path> directory for old tpm files;\n";
    print "\n";		       
    print "    Packages :\n";
    print "    --name=<patter>         process only the files matching the pattern;\n";
    print "    --all                   process all package files;\n";
    print "    --arch=<(all|name)>     process some or all architectures.\n";
    print "\n";
    print "    Support commands:\n";
    print "    --log=<file name>       log activity to file;\n";
    print "    --debug                 turn on debug trace;\n";
    print "    --verbose               some more messages\n";
    print "    --help                  displays this help text and exits\n";
}

sub build_zip {
  local ($subdir, $tpmname) = @_;
  Tpm::Tpm2Zip("$opt_tpm_dir/$subdir/$tpmname.tpm", $opt_tex_dir, "$opt_dest_dir/$subdir", $opt_arch);
}

sub read_dependencies {
  local ($subdir, $tpmname) = @_;
  @{$req{"$subdir/$tpmname"}} = &Tpm::TpmGetRequires("$opt_tpm_dir/$subdir/$tpmname.tpm");
}

sub read_files {
  local ($subdir, $tpmname) = @_;
  return Tpm::TpmGetFiles("$opt_tpm_dir/$subdir/$tpmname.tpm", $opt_arch);
}

sub collect_tpmfiles {
  local ($subdir, $tpmname) = @_;
  push @tpm_files, "$subdir/$tpmname";
}

sub do_subdir {
  local ($subdir, $fn) = @_;
#  print "subdir = $subdir\n";
  while (<$subdir/*>) {
    local $tpmname = $_;
    next if ($tpmname =~ m/~$/);
    $tpmname =~ m/.*\/([^\/]*)\/([^\/]*)\.tpm$/;
    $subdir = $1;
    $tpmname = $2;
    print "Doing $tpmname\n" if ($opt_verbose);
    &{$fn}($subdir, $tpmname);
  }
}

#
# Check that all tpm package is required by another one at least
#
sub check_dependencies {
  local %req, @tpm_files;
  local $s, $k, @l;

  for $s (@Tpm::TpmSubdirs) {
    $opt_subdir = $s;
    if (! -d $opt_subdir) {
      if (! -d "$opt_tpm_dir/$opt_subdir") {
	printf STDERR "$0: $opt_subdir is not a tpm subdirectory, skipping!\n";
#	exit 1;
	next;
      }
      else {
	$opt_subdir = "$opt_tpm_dir/$opt_subdir";
      }
    }
    &do_subdir($opt_subdir, \&read_dependencies);
    &do_subdir($opt_subdir, \&collect_tpmfiles);
  }

  @l = ("collections/texlive");
  while ($#l >= 0) {
    my $p = pop @l;
    remove_list(\@tpm_files, $p);
    push @l, @{$req{$p}};
  }
  print "***********************************\n";
  print "Packages not required by anything :\n";
  print "***********************************\n";
  map { print "$_\n"; } @tpm_files;
  print "\n";
  print "***********************************\n";
}

sub collect_files {
  local ($d, @lf) = @_;
  my ($f);
  foreach $f (@lf) {
    my $path = "$d/$f";
    if (-f "$path") {
      $path =~ s@^${opt_tex_dir}/@@;
      $path =~ s@\\@/@g;
      if ($path !~ m@^texmf/tpm/lists/@) { push @texmf_files, $path; }
    }
  }
}

sub collect_coverage {
  local ($subdir, $tpmname) = @_;
  local (@l) = &read_files($subdir, $tpmname);
  print "Pushing $#l files from $subdir/$tpmname\n" if $opt_debug;
  push @tpm_files, @l;
}

sub find_multiple {
  local (*files) = @_;
  my (@l) = sort(@files);
  my(@res, $e1, $e2);
  @files = ( );
  $e1 = pop(@l);
  while ($e1) {
    push @files, $e1;
    while (($e2 = pop(@l)) eq $e1) {
      push @res, $e2;
    }
    $e1 = $e2;
  }
  return @res;
}

#
# Check that all files in tpm files exist
# and that all files in the texmf tree
# are requested by one tpm file.
#
sub check_coverage {
  local (@texmf_files, @tpm_files);
  my($s);

  &walk_dir("$opt_tex_dir/texmf", \&collect_files);
  &walk_dir("$opt_tex_dir/bin", \&collect_files);
  &walk_dir("$opt_tex_dir/info", \&collect_files);
  &walk_dir("$opt_tex_dir/man", \&collect_files);
  &walk_dir("$opt_tex_dir/Books", \&collect_files);
  &walk_dir("$opt_tex_dir/usergrps", \&collect_files);

  for $s (@Tpm::TpmSubdirs) {
    $opt_subdir = $s;
    if (! -d $opt_subdir) {
      if (! -d "$opt_tpm_dir/$opt_subdir") {
	printf STDERR "$0: $opt_subdir is not a tpm subdirectory, skipping!\n";
#	exit 1;
	next;
      }
      else {
	$opt_subdir = "$opt_tpm_dir/$opt_subdir";
      }
    }
    &do_subdir($opt_subdir, \&collect_coverage);
  }
  print "Found $#tpm_files files in tpm files.\n";
  local (@doublons) = find_multiple(\@tpm_files);
  print "*****************************************\n";
  print "Files found multiple times in tpm files :\n";
  map { print "$_\n"; } @doublons;
  print "*****************************************\n";
  local (@l1, @l2);
  &diff_list(\@texmf_files, \@tpm_files, \@l1, \@l2);
  print "*****************************************\n";
  print "Files in texmf tree not in tpm files :\n";
  map { print "$_\n"; } @l1;
  print "*****************************************\n";
  print "Files in tpm files not in texmf tree :\n";
  print "*****************************************\n";
  map { print "$_\n"; } @l2;
  print "*****************************************\n";

}

sub main {

  unless (&NGetOpt ("tpm2zip", "pkg2tpm", "tpm_import",
		    "pkg_dir=s", "tpm_dir=s", "texdir=s",
		    "import_tpm_dir=s", "name=s", "arch=s",
		    "check=s", "tpm_index",
		    "all", "log=s",
		    "verbose", "debug", "help"))
    {
	print STDERR "Try `$0 --help'";
	exit 1;
    }

  if ($opt_help) {
    &help;
    exit 0;
  }

  if (($opt_tpm2pkg && $opt_pkg2tpm)
     || ($opt_tpm2pkg && $opt_tpm_import)
     || ($opt_pkg2tpm && $opt_tpm_import)) {
    print STDERR "$0: `--tpm2pkg', `--pkg2tpm' and `--tpm_import' are incompatible.\n";
    exit 1;
  }

  if ($opt_name && $opt_all) {
    print STDERR "$0: `--all' and `--name' are incompatible.\n";
    exit 1;
  }

  if ($opt_arch 
      && !(($opt_arch eq "all")
	   || &member($opt_arch, @Tpm::arch_list))) {
    print STDERR "$0: $opt_arch is neither `all' nor a valid architecture, aborting.\n";
    exit 1;
  }

  if ($opt_log) {
#    $logfile = "$topdir\\win32\\$opt_log";
    if (! $opt_log =~ /\.log$/) {
      $opt_log .= ".log";
    }
    # start redirection if asked
    &start_redirection($opt_log);
 }

  if ($opt_tpm_import) {
    my $pkgname = "";
    if ($opt_name) {
      print "${opt_import_tpm_dir}/${opt_name}.tpm\n";
      while (<${opt_import_tpm_dir}/${opt_name}.tpm>) {
        my $name = $_;
	print "name = $name\n";
	$name =~ m@^${opt_import_tpm_dir}[/\\]([^/\\]*)[/\\](.*)\.tpm$@;
        my $dst_subdir = $1;
        for my $key (keys %srcdst) {
	  if ($srcdst{$key} eq $dst_subdir) {
	    $src_subdir = $key;
	    last;
	  }
	}
        $name = $2;
	$pkgname = "$name.pkg";
	Tpm::OldTpm2Pkg("$opt_import_tpm_dir/$dst_subdir/$name.tpm",
			$pkgname, $opt_tex_dir, "$opt_pkg_dir/$src_subdir");
      }
      if ($pkgname eq "") {
	print STDERR "$0: $name.tpm not found !\n";
	$retcode = 1;
      }
    }
    elsif ($opt_all) {
      foreach $src_subdir (keys %srcdst) {
	my $dst_subdir = $srcdst{$src_subdir};
	while (<$opt_import_tpm_dir/$dst_subdir/*.tpm>) {
	  $pkgname = $_;
	  $pkgname =~ s@^.*/([^/]*)\.tpm$@\1.pkg@;
	  print "processing $_\n\t-> $src_subdir/$pkgname\n";
	  Tpm::OldTpm2Pkg($_, $pkgname, $opt_tex_dir, "$opt_pkg_dir/$src_subdir");
	}
      }
    }
  }

  elsif ($opt_pkg2tpm) {
    my $tpmname = "";
    if ($opt_name) {
      while (<${opt_pkg_dir}/${opt_name}.pkg>) {
        my $name = $_;
	$name =~ m@^${opt_pkg_dir}[/\\]([^\//]*)[/\\](.*)\.pkg$@;
        my $src_subdir = $1;
        $name = $2;
        $tpmname = "$name.tpm";
	Tpm::Pkg2Tpm($tpmname, "$opt_pkg_dir/$src_subdir/$name.pkg", 
		     $opt_tex_dir, "$opt_tpm_dir/" . $srcdst{$src_subdir});
      }
      if ($tpmname eq "") {
	print STDERR "$0: $name.pkg not found !\n";
	$retcode = 1;
      }
    }
    elsif ($opt_all) {
      foreach $subdir (keys %srcdst) {
	while (<$opt_pkg_dir/$subdir/*.pkg>) {
	  $tpmname = $_;
	  my $destdir = $srcdst{$subdir};
#	  next if (/-static/);
	  $tpmname =~ s@^.*/([^/]*)\.pkg$@\1.tpm@;
	  print "processing $tpmname -> $destdir/$tpmname\n";
	  Tpm::Pkg2Tpm( $tpmname, $_, $opt_tex_dir, "$opt_tpm_dir/$destdir");
	  open IN, "<$opt_tpm_dir/$destdir/$tpmname";
	  my $s = "";
	  while (<IN>) {
	    next if ($_ =~ /^\s*$/);
	    $s .= $_;
	  }
	  close IN;
	  open OUT, ">$opt_tpm_dir/$destdir/$tpmname.new";
	  $s =~ s@(\s*)<(TPM:[^>]*)>\n([^\n]*)\n\s*</\2>@\1<\2>\3</\2>@gms;
	  print OUT $s;
	  close OUT;
	  rename ("$opt_tpm_dir/$destdir/$tpmname.new", "$opt_tpm_dir/$destdir/$tpmname");
	}
      }
    }
  }

  elsif ($opt_tpm2zip) {
    if ($opt_all) {
      for my $s (@Tpm::TpmSubdirs) {
	next if ($s =~ /schemes/i);
	if (! -d "$opt_tpm_dir/$s") {
	  printf STDERR "$0: $s is not a tpm subdirectory, skipping!\n";
	  next;
	} 
	else {
	  $opt_subdir = "$opt_tpm_dir/$s";
	}
	print "subdir = $opt_subdir\n";
	&do_subdir($opt_subdir, \&build_zip);
      }
    } 
#      elsif ($opt_subdir) {
#        if (! -d $opt_subdir) {
#  	if (! -d "$opt_tpm_dir/$opt_subdir") {
#  	  printf STDERR "$0: $opt_subdir is not a tpm subdirectory, aborting!\n";
#  	  exit 1;
#  	} else {
#  	  $opt_subdir = "$opt_tpm_dir/$opt_subdir";
#  	}
#        }
#        &do_subdir($opt_subdir, \&build_zip);
#      } 
    elsif ($opt_name) {
      while (<"${opt_name}">) {
        my $name = $_;
	$name .= ".tpm" if ($name !~ m/\.tpm$/);
	if (! -f $name) {
	  if (! -f "$opt_tpm_dir/$name") {
	    printf STDERR "$0: $opt_name non existent tpm file, aborting!\n";
	    exit 1;
	  }
	}
	$name =~ m/([^\/]*)\/(.*)$/;
	my $subdir = $1;
	if (! -d "$opt_dest_dir/$subdir") {
	  mkdir "$opt_dest_dir/$subdir";
	}
	Tpm::Tpm2Zip("$opt_tpm_dir/$name", $opt_tex_dir, "$opt_dest_dir/$subdir", $opt_arch);
      }
    }
  }
  elsif ($opt_tpm_index) {
    #
    # Rebuild tpm.zip file
    #
    unlink("${opt_dest_dir}/tpm.zip");
    $cwd = &getcwd();
    chdir($opt_tex_dir);
    open ZIP, "| zip -9\@or ${opt_dest_dir}/tpm.zip  > nul";
    foreach $e (@Tpm::TpmSubdirs) {
      while (<texmf/tpm/$e/*>) {
	print ZIP "$_\n" if (/\.tpm$/io);
	print  "$_\n" if (/\.tpm$/io and $opt_verbose);
      }
    }
    close ZIP;
    chdir($cwd);
    print "File `${opt_dest_dir}/tpm.zip' rebuilt.\n";
  }
  # Check coverage of packages: do the tpm files cover
  # the list of files in the texmf tree?
  elsif ($opt_check && $opt_check =~ m/^cov/) {
    print "Checking coverage:\n";
    $opt_arch = "all";
    &check_coverage;
  }
  # Check dependencies: are all the tpm packages required
  # by some other package?
  elsif ($opt_check && $opt_check =~ m/^dep/) {
    print "Checking dependencies:\n";
    &check_dependencies;
  }


  if ($opt_log) {
    &stop_redirection($opt_log)
  }
}
